"""capdata Provider Module."""
from capinvest_core.provider.abstract.provider import Provider
from capinvest_capdata.models.equity_historical import CapdataEquityHistoricalFetcher
from capinvest_capdata.models.etf_historical import CapdataEtfHistoricalFetcher
from capinvest_capdata.models.futures_historical import CapdataFuturesHistoricalFetcher
from capinvest_capdata.models.index_historical import CapdataIndexHistoricalFetcher
from capinvest_capdata.models.yield_curve import CapdataYieldCurveFetcher

capdata_provider = Provider(
    name="capdata",
    website="https://www.caprisktech.com/",
    description="通过获取及时和全面的市场数据和深度业务分析数据，在讯息万变的金融市场中做出基于风险的投资与交易决策。CapData提供为您定制的市场数据、模型数据和合成数据等",
    credentials=["api_key"],  # api_key OR (user_name@_@user_pwd,针对使用用户名和密码的使用@_@分割)
    fetcher_dict={
        "EquityHistorical": CapdataEquityHistoricalFetcher,
        "FuturesHistorical": CapdataFuturesHistoricalFetcher,
        "EtfHistorical": CapdataEtfHistoricalFetcher,
        "YieldCurve": CapdataYieldCurveFetcher,
        "IndexHistorical": CapdataIndexHistoricalFetcher,
    },
    repr_name="CapData Public Data API",
    instructions="Sign up for a free API key here: https://www.caprisktech.com/#/documents/http-documentation",
)



"""Capdata Yield Curve Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal
from typing import Optional

from capinvest_capdata.utils.data_path import bond_yield_curve_url, \
    ir_yield_curve_url
from capinvest_capdata.utils.helpers import post_by_token
from capinvest_core.provider.abstract.fetcher import Fetcher
from capinvest_core.provider.standard_models.yield_curve import (
    YieldCurveData,
    YieldCurveQueryParams,
)
from capinvest_core.provider.utils.errors import EmptyDataError
from pydantic import Field

from capinvest_capdata.utils.helpers import get_capdata_token


class CapdataYieldCurveQueryParams(YieldCurveQueryParams):
    """Capdata Yield Curve Query Params."""
    __json_schema_extra__ = {"date": {"multiple_items_allowed": False}}
    curve_code: str = Field(
        default="CN_TREAS_STD",
        description="curve code",
    )
    yield_curve_type: Literal["BOND", "IR"] = (
        Field(
            default="BOND",
            description="The yield curve type.",
        )
    )


class CapdataYieldCurveData(YieldCurveData):
    """Capdata Yield Curve Data."""


class CapdataYieldCurveFetcher(
    Fetcher[
        CapdataYieldCurveQueryParams,
        list[CapdataYieldCurveData],
    ]
):
    """Capdata Yield Curve Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> CapdataYieldCurveQueryParams:
        """Transform query."""
        transformed_params = params
        if params.get("date") is None:
            transformed_params["date"] = datetime.now().date()
        return CapdataYieldCurveQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
            query: CapdataYieldCurveQueryParams,
            credentials: Optional[dict[str, str]],
            **kwargs: Any,
    ) -> list[dict]:
        """Extract data."""
        # pylint: disable=import-outside-toplevel
        params = {
            "curve": query.curve_code,
            "start": query.date,
            "end": query.date,
            "freq": "d",
            "parseProto": True
        }
        if query.yield_curve_type == "BOND":
            results = post_by_token(bond_yield_curve_url, params, credentials)
        else:
            results = post_by_token(ir_yield_curve_url, params, credentials)
        if not results:
            raise EmptyDataError(
                f"{query.curve_code} No data found"
            )
        return results

    @staticmethod
    def transform_data(
            query: CapdataYieldCurveQueryParams,
            data: list[dict],
            **kwargs: Any,
    ) -> list[CapdataYieldCurveData]:
        """Transform data."""
        # pylint: disable=import-outside-toplevel
        if not data:
            raise EmptyDataError("The request was returned empty.")
        curve_data = data[0]['curve']['curve']
        date_list = curve_data['pillarDate']
        maturity_list = curve_data['pillarName']
        pillar_values = curve_data['pillarValues']['data']
        records = []
        for i in range(len(date_list)):
            records.append({
                "date": date_list[i],
                "maturity": maturity_list[i],
                "value": pillar_values[i]
            })
        return [CapdataYieldCurveData.model_validate(d) for d in records]



"""Deribit Futures Historical Model."""
from datetime import datetime
from typing import Any, Optional, List, Dict

from capinvest_capdata.utils.data_path import history_url
from capinvest_capdata.utils.helpers import CAPDATA_INTERVALS, CapdataIntervals, post_by_token
from capinvest_core.provider.abstract.fetcher import Fetcher
from capinvest_core.provider.standard_models.equity_historical import EquityHistoricalQueryParams, EquityHistoricalData
from capinvest_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from capinvest_core.provider.utils.errors import EmptyDataError
from dateutil.relativedelta import relativedelta
from pydantic import Field

from capinvest_capdata.utils.helpers import get_capdata_token


# pylint: disable=unused-argument


class CapdataEtfHistoricalQueryParams(EquityHistoricalQueryParams):
    """
    Capdata etf historical Price Query.

    Source: https://www.caprisktech.com/#/documents/http-documentation
    """

    __alias_dict__ = {"start_date": "from", "end_date": "to"}
    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
        "interval": {"multiple_items_allowed": False,
                     "choices": CAPDATA_INTERVALS},
    }

    interval: CapdataIntervals = Field(
        default="d", description=QUERY_DESCRIPTIONS.get("interval", "")
    )


class CapdataEtfHistoricalData(EquityHistoricalData):
    """Capdata etf Historical Data."""


class CapdataEtfHistoricalFetcher(
    Fetcher[CapdataEtfHistoricalQueryParams, list[CapdataEtfHistoricalData]]
):
    """Deribit etf Historical Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> CapdataEtfHistoricalQueryParams:
        """Transform the query."""
        transformed_params = params
        now = datetime.now().date()
        if params.get("start_date") is None:
            if params.get("interval") == "d" or params.get("interval") == "w":
                transformed_params["start_date"] = now - relativedelta(years=1)
            else:
                transformed_params["start_date"] = now - relativedelta(day=10)
        if params.get("end_date") is None:
            transformed_params["end_date"] = now
        return CapdataEtfHistoricalQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
            query: CapdataEtfHistoricalQueryParams,
            credentials: Optional[dict[str, str]],
            **kwargs: Any,
    ) -> list:
        """Extract the raw data."""
        """Return the raw data from the CapData endpoint."""
        symbols = query.symbol.split(",")
        params = {
            "inst": symbols,
            "start": query.start_date.strftime("%Y-%m-%d %H:%M:%S"),
            "end": query.end_date.strftime("%Y-%m-%d %H:%M:%S"),
            "freq": query.interval,
            "fields": ["volume", "open", "close", "low", "high"]
        }
        results = post_by_token(history_url, params, credentials)
        if not results:
            raise EmptyDataError(
                f"{query.symbol} No data found"
            )
        return results

    @staticmethod
    def transform_data(
            query: CapdataEtfHistoricalQueryParams, data: List[Dict], **kwargs: Any
    ) -> list[CapdataEtfHistoricalData]:
        """Transform the data."""
        results: List[CapdataEtfHistoricalData] = []
        for d in sorted(
                data,
                key=lambda x: (
                        (x["date"], x["inst"])
                        if len(query.symbol.split(",")) > 1
                        else x["date"]
                ),
                reverse=False,
        ):
            results.append(CapdataEtfHistoricalData.model_validate(d))
        return results



import datetime
import json
import logging
from typing import Literal

import requests
from cachetools import TTLCache

from capinvest_capdata.utils.data_path import auth_url

CAPDATA_INTERVALS = ["1m", "15m", "30m", "1h", "d", "w"]
CapdataIntervals = Literal["1m", "15m", "30m", "1h", "d", "w"]
logger = logging.getLogger("uvicorn.info")
capdata_user_api_key_cache = TTLCache(maxsize=1000, ttl=60 * 3)


def post_by_token(url, params, credentials: dict):
    token = get_capdata_token(credentials)
    logger.info(f"{datetime.datetime.now()} capdata url: {url}")
    headers = {'Accept': 'application/json', 'content-type': 'application/json', 'Accept-Encoding': 'gzip',
               'x-access-token': token}
    response = requests.post(url, json=params, headers=headers).text
    res = json.loads(response)
    if 'respCode' not in res.keys():
        raise Exception(res["error"])
    code = res["respCode"]
    if str(code) == 'SUCCEED':
        if 'data' in res:
            return res["data"]
        else:
            return None
    else:
        raise Exception(res["message"])


def get_capdata_token(credentials: dict):
    """Get CapData token using API key or username/password.
    
    Supports two authentication methods:
    1. Using capdata_api_key directly
    """
    # Check if api_key is provided and not empty
    api_key = credentials.get("capdata_api_key") if credentials else None
    if api_key and api_key.strip():
        if "@_@" not in api_key:
            return api_key
        else:
            if api_key in capdata_user_api_key_cache:
                return capdata_user_api_key_cache.get(api_key)
            api_keys = api_key.split("@_@")
            capdata_user_name = api_keys[0]
            capdata_user_pwd = api_keys[1]
            header = {'Accept': 'application/json', 'content-type': 'application/json', 'Accept-Encoding': 'gzip'}
            response = requests.post(auth_url,
                                     json={"account": capdata_user_name,
                                           "pwd": capdata_user_pwd},
                                     headers=header).text
            res = json.loads(response)
            if 'respCode' not in res.keys():
                raise Exception(res["error"])
            code = res["respCode"]
            if str(code) == 'SUCCEED':
                token = res["data"]
                capdata_user_api_key_cache[api_key] = token
                return token
            else:
                raise Exception(res["message"])
    raise PermissionError(
        "Capdata authentication failed: Please provide either capdata_api_key(如果是账号密码格式为:capdata_user_name@_@capdata_user_pwd)")



# capdata 认证路径
auth_url = "https://www.capdata.pro/capdata/auth"
# 历史行情路径
history_url = "https://www.capdata.pro/capdata/get/hist/mkt"
# 债券收益率曲线数据路径
bond_yield_curve_url = "https://www.capdata.pro/capdata/get/bond/curve"
# 利率收益率曲线数据路径
ir_yield_curve_url = "https://www.capdata.pro/capdata/get/ir/curve"
